"""ControlledProcess classes that periodically also produce messages to heartbeat topics
and/or self-produces own logs
"""

# imports
import datetime, inspect
from abc import ABC
from confluent_kafka.serialization import StringSerializer
from confluent_kafka import SerializingProducer
from openmsitoolbox.argument_parsing.has_arguments import HasArguments
from openmsitoolbox.controlled_process.controlled_process import ControlledProcess
from openmsitoolbox import ControlledProcessSingleThread, ControlledProcessMultiThreaded
from ..kafka_wrapper.config_file_parser import KafkaConfigFileParser
from ..kafka_wrapper import OpenMSIStreamProducer
from .heartbeat_producibles import HeartbeatProducible
from .log_producibles import LogProducible
from .log_handler import LoggingHandler


class ControlledProcessHeartbeatsLogs(ControlledProcess, HasArguments, ABC):
    """A long-running process that occasionally produces messages
    to a heartbeat topic and/or self-produces own logs
    """

    def __init__(
        self,
        config_path,
        *args,
        heartbeat_topic_name=None,
        heartbeat_program_id=None,
        heartbeat_interval_secs=None,
        log_topic_name=None,
        log_program_id=None,
        log_interval_secs=None,
        **kwargs,
    ):
        try:
            super().__init__(config_path, *args, **kwargs)
        except TypeError:
            super().__init__(*args, **kwargs)
        self.__heartbeat_topic_name = heartbeat_topic_name
        self.__default_heartbeat_program_id = heartbeat_program_id
        self._heartbeat_program_id = None
        self.__heartbeat_interval_secs = heartbeat_interval_secs
        self.__heartbeat_producer = None
        self.__heartbeat_producer_owned = False
        self.__last_heartbeat = datetime.datetime.fromtimestamp(0)
        self.__log_topic_name = log_topic_name
        self.__default_log_program_id = log_program_id
        self._log_program_id = None
        self.__log_interval_secs = log_interval_secs
        self.__log_producer = None
        self.__log_producer_owned = False
        self.__last_log = datetime.datetime.fromtimestamp(0)
        # Read config file for later use
        self.__config_path = config_path
        cfp = KafkaConfigFileParser(config_path, logger=self.logger)
        if "heartbeat" in cfp.available_group_names:
            self.__heartbeat_config_dict = cfp.heartbeat_configs
        else:
            self.__heartbeat_config_dict = {}
        if "log" in cfp.available_group_names:
            self.__log_config_dict = cfp.log_configs
        else:
            self.__log_config_dict = {}
        self.__broker_configs = cfp.broker_configs

    def set_heartbeat_producer(self, producer, close_it=False):
        "Set producer instance for generating heartbeats"
        if self.__heartbeat_topic_name is None:
            # Nothing to do
            producer = None
        # Get configuration preference
        prefer_separate_producer = False
        if (
            "prefer.separate.producer" in self.__heartbeat_config_dict
            and self.__heartbeat_config_dict["prefer.separate.producer"]
        ):
            prefer_separate_producer = True
        if producer == "Separate":
            prefer_separate_producer = True
        # Order of priority:
        # 1. If producer is None, turn off heartbeats.
        # 2. If prefer_separate_producer is False, and producer is a valid
        #    producer object, use that.
        # 3. Otherwise, use our own producer.
        if producer is None:
            self.__set_heartbeat_producer_internal(None, False)
        elif (
            not prefer_separate_producer
            and hasattr(producer, "produce_object")
            and inspect.isroutine(producer.produce_object)
            and hasattr(producer, "producer_id")
        ):
            self.__set_heartbeat_producer_internal(producer, close_it)
        else:
            own_producer = self.__get_standalone_heartbeat_producer()
            self.__set_heartbeat_producer_internal(own_producer, True)

    def __get_standalone_heartbeat_producer(self):
        "Get a standalone producer for heartbeats"
        if len(self.__heartbeat_config_dict) < 1:
            self.logger.warning(
                (
                    f"WARNING: config file at {self.__config_path} has no 'heartbeat' section but "
                    "a heartbeat topic name was given and a standalone producer is in use."
                ),
            )
        heartbeat_config_dict = self.__heartbeat_config_dict.copy()
        # Remove non-broker parameters
        if "prefer.separate.producer" in heartbeat_config_dict:
            del heartbeat_config_dict["prefer.separate.producer"]
        # Add broker parameters
        if "key.serializer" not in heartbeat_config_dict:
            heartbeat_config_dict["key.serializer"] = StringSerializer()
        if "value.serializer" not in heartbeat_config_dict:
            heartbeat_config_dict["value.serializer"] = StringSerializer()
        all_heartbeat_producer_configs = {}
        if "bootstrap.servers" not in heartbeat_config_dict:
            all_heartbeat_producer_configs.update(self.__broker_configs)
        all_heartbeat_producer_configs.update(heartbeat_config_dict)
        return OpenMSIStreamProducer(
            SerializingProducer, all_heartbeat_producer_configs, logger=self.logger
        )

    def __set_heartbeat_producer_internal(self, producer, owned):
        "Set/update current heartbeat producer, cleaning up old as required"
        if (
            self.__heartbeat_producer == producer
            and self.__heartbeat_producer_owned == owned
        ):
            # no change
            return
        # Cleanup old one if present
        if self.__heartbeat_producer is not None:
            # send a final heartbeat
            self.__heartbeat_producer.produce_object(
                self.get_heartbeat_message(), self.__heartbeat_topic_name
            )
            self.logger.info("Cleaning up heartbeat producer")
            if (
                self.__heartbeat_producer == self.__log_producer
                and self.__heartbeat_producer_owned
            ):
                self.__log_producer_owned = True  # Hand off ownership
                self.__heartbeat_producer_owned = False
            if self.__heartbeat_producer_owned:
                self.__heartbeat_producer.flush()
                self.__heartbeat_producer.close()
            self.__heartbeat_producer = None
            self.__heartbeat_producer_owned = None
        # Set new one
        self.__heartbeat_producer = producer
        self.__heartbeat_producer_owned = owned
        if self.__heartbeat_producer is not None:
            self.logger.info(
                (
                    "Heartbeat messages will be produced to the "
                    f"'{self.__heartbeat_topic_name}' topic every "
                    f"{self.__heartbeat_interval_secs} seconds"
                )
            )
            self._heartbeat_program_id = self.__default_heartbeat_program_id
            if self._heartbeat_program_id is None:
                self._heartbeat_program_id = self.__heartbeat_producer.producer_id

    def set_log_producer(self, producer, close_it=True):
        "Set producer instance for generating logs"
        if self.__log_topic_name is None:
            # Nothing to do
            producer = None
        # Get configuration preference
        prefer_separate_producer = False
        if (
            "prefer.separate.producer" in self.__log_config_dict
            and self.__log_config_dict["prefer.separate.producer"]
        ):
            prefer_separate_producer = True
        if producer == "Separate":
            prefer_separate_producer = True
        # Order of priority:
        # 1. If producer is None, turn off logs.
        # 2. If prefer_separate_producer is False, and producer is a valid
        #    producer object, use that.
        # 3. Otherwise, use our own producer.
        if producer is None:
            self.__set_log_producer_internal(None, False)
        elif (
            not prefer_separate_producer
            and hasattr(producer, "produce_object")
            and inspect.isroutine(producer.produce_object)
            and hasattr(producer, "producer_id")
        ):
            self.__set_log_producer_internal(producer, close_it)
        else:
            own_producer = self.__get_standalone_log_producer()
            self.__set_log_producer_internal(own_producer, True)

    def __get_standalone_log_producer(self):
        "Get a standalone producer for logs"
        if len(self.__log_config_dict) < 1:
            self.logger.warning(
                (
                    f"WARNING: config file at {self.__config_path} has no 'log' section but "
                    "a log topic name was given and a standalone producer is in use."
                ),
            )
        log_config_dict = self.__log_config_dict.copy()
        # Remove non-broker parameters
        if "prefer.separate.producer" in log_config_dict:
            del log_config_dict["prefer.separate.producer"]
        if "max.logs.per.message" in log_config_dict:
            del log_config_dict["max.logs.per.message"]
        if "log.level" in log_config_dict:
            del log_config_dict["log.level"]
        # Add broker parameters
        if "key.serializer" not in log_config_dict:
            log_config_dict["key.serializer"] = StringSerializer()
        if "value.serializer" not in log_config_dict:
            log_config_dict["value.serializer"] = StringSerializer()
        all_log_producer_configs = {}
        if "bootstrap.servers" not in log_config_dict:
            all_log_producer_configs.update(self.__broker_configs)
        all_log_producer_configs.update(log_config_dict)
        return OpenMSIStreamProducer(
            SerializingProducer, all_log_producer_configs, logger=self.logger
        )

    def __set_log_producer_internal(self, producer, owned):
        "Set/update current log producer, cleaning up old as required"
        if self.__log_producer == producer and self.__log_producer_owned == owned:
            # no change
            return
        # Cleanup old one if present
        if self.__log_producer is not None:
            # send a final log
            self.__log_producer.produce_object(
                self.get_log_message(), self.__log_topic_name
            )
            self.logger.info("Cleaning up log producer")
            if (
                self.__heartbeat_producer == self.__log_producer
                and self.__log_producer_owned
            ):
                self.__heartbeat_producer_owned = True  # Hand off ownership
                self.__log_producer_owned = False
            if self.__log_producer_owned:
                self.__log_producer.flush()
                self.__log_producer.close()
            self.__log_producer = None
            self.__log_producer_owned = None
        # Set new one
        self.__log_producer = producer
        self.__log_producer_owned = owned
        if self.__log_producer is not None:
            self.logger.info(
                (
                    "log messages will be produced to the "
                    f"'{self.__log_topic_name}' topic every "
                    f"{self.__log_interval_secs} seconds"
                )
            )
            self._log_program_id = self.__default_log_program_id
            if self._log_program_id is None:
                self._log_program_id = self.__log_producer.producer_id
            max_messages = 65536
            if "max.logs.per.message" in self.__log_config_dict:
                max_messages = int(self.__log_config_dict["max.logs.per.message"])
            LoggingHandler.set_max_messages(max_messages)
            if "log.level" in self.__log_config_dict:
                LoggingHandler.setLevel(self.__log_config_dict["log.level"])

    def get_heartbeat_message(self):
        """Return the HeartbeatProducible-type object that should be produced to the
        heartbeat topic
        """
        return HeartbeatProducible(self._heartbeat_program_id)

    def get_log_message(self):
        """Return the LogProducible-type object that should be produced to the
        log topic
        """
        return LogProducible(self._log_program_id)

    def _print_still_alive(self):
        """Print the "still alive" character to the console like a regular
        ControlledProcess, but also produce messages to the heartbeat and/or
        log topics if applicable
        """
        super()._print_still_alive()
        if (
            self.__heartbeat_producer is not None
            and (datetime.datetime.now() - self.__last_heartbeat).total_seconds()
            > self.__heartbeat_interval_secs
        ):
            self.__heartbeat_producer.produce_object(
                self.get_heartbeat_message(), self.__heartbeat_topic_name
            )
            self.__last_heartbeat = datetime.datetime.now()
        if (
            self.__log_producer is not None
            and (datetime.datetime.now() - self.__last_log).total_seconds()
            > self.__log_interval_secs
        ):
            self.__log_producer.produce_object(
                self.get_log_message(), self.__log_topic_name
            )
            self.__last_log = datetime.datetime.now()

    def _on_shutdown(self):
        try:
            super()._on_shutdown()
        except NotImplementedError:
            pass
        # Cleanup
        self.__set_heartbeat_producer_internal(None, False)
        self.__set_log_producer_internal(None, False)

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "heartbeat_topic_name",
            "heartbeat_program_id",
            "heartbeat_interval_secs",
            "log_topic_name",
            "log_program_id",
            "log_interval_secs",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        kwargs = {
            **superkwargs,
            "heartbeat_topic_name": parsed_args.heartbeat_topic_name,
            "heartbeat_program_id": parsed_args.heartbeat_program_id,
            "heartbeat_interval_secs": parsed_args.heartbeat_interval_secs,
            "log_topic_name": parsed_args.log_topic_name,
            "log_program_id": parsed_args.log_program_id,
            "log_interval_secs": parsed_args.log_interval_secs,
        }
        return superargs, kwargs


class ControlledProcessSingleThreadHeartbeatsLogs(
    ControlledProcessHeartbeatsLogs, ControlledProcessSingleThread
):
    """A long-running process with a single worker thread that occasionally produces
    messages to a heartbeat and/or log topic
    """


class ControlledProcessMultiThreadedHeartbeatsLogs(
    ControlledProcessHeartbeatsLogs, ControlledProcessMultiThreaded
):
    """A long-running process with multiple worker threads that occasionally produces
    messages to a heartbeat and/or log topic
    """



"""A set of Consumers and Producers that share share the same KafkaCrypto key-passing instance"""

# imports
from openmsitoolbox import LogOwner
from .consumer_group import ConsumerGroup
from .producer_group import ProducerGroup


class ConsumerAndProducerGroup(LogOwner):
    """
    Class for working with a group of Consumers and Producers sharing a single
    :class:`kafkacrypto.KafkaCrypto` instance

    :param config_path: Path to the config file that should be used to define
        Consumers/Producers in the group
    :type config_path: :class:`pathlib.Path`
    :param consumer_topic_name: The name of the topic to which the Consumers should be subscribed
    :type consumer_topic_name: str
    :param consumer_group_id: The ID string that should be used for each Consumer in the group.
        "create_new" (the defaults) will create a new UID to use.
    :type consumer_group_id: str, optional
    :param kafkacrypto: The :class:`~OpenMSIStreamKafkaCrypto` object that should be used
        to instantiate Consumers. Only needed if a single specific
        :class:`~OpenMSIStreamKafkaCrypto` instance should be shared.
    :type kafkacrypto: :class:`~OpenMSIStreamKafkaCrypto`, optional
    :param treat_undecryptable_as_plaintext: If True, the KafkaCrypto Deserializers
        will immediately return any keys/values that are not possibly decryptable as
        binary data. This allows faster handling of messages that will never be
        decryptable, such as when enabling or disabling encryption across a platform,
        or when unencrypted messages are mixed in a topic with encrypted messages.
    :type treat_undecryptable_as_plaintext: boolean, optional
    :param max_wait_per_decrypt: Number of seconds a KafkaCrypto Deserializer
        waits before giving up.
    :type max_wait_per_decrypt: float, optional
    :param max_initial_wait_per_decrypt: Number of seconds a KafkaCrypto Deserializer
        waits before giving up on the first message it sees.
    :type max_initial_wait_per_decrypt: float, optional
    """

    @property
    def kafkacrypto(self):
        """
        The KafkaCrypto object handling key passing and serdes for encrypted messages
        """
        return self.__kafkacrypto

    @property
    def consumer_topic_name(self):
        """
        Name of the topic to which the consumers are subscribed
        """
        if self.__consumer_group is not None:
            return self.__consumer_group.consumer_topic_name
        return None

    @property
    def consumer_group_id(self):
        """
        Group ID of all consumers in the group
        """
        if self.__consumer_group is not None:
            return self.__consumer_group.consumer_group_id
        return None

    def __init__(
        self,
        config_path,
        *,
        consumer_topic_name=None,
        consumer_group_id="create_new",
        kafkacrypto=None,
        treat_undecryptable_as_plaintext=False,
        max_wait_per_decrypt=5.0,
        max_initial_wait_per_decrypt=60.0,
        **kwargs,
    ):
        """
        Constructor method
        """
        # must remove producer_topic_name since we define it as
        # part of this class even though we do not use it in init.
        other_kwargs = kwargs.copy()
        if "producer_topic_name" in other_kwargs:
            del other_kwargs["producer_topic_name"]
        super().__init__(**other_kwargs)
        self.__producer_group = ProducerGroup(
            config_path, kafkacrypto=kafkacrypto, logger=self.logger
        )
        self.__kafkacrypto = self.__producer_group.kafkacrypto
        self.__consumer_group = None
        if consumer_topic_name is not None:
            self.__consumer_group = ConsumerGroup(
                config_path,
                consumer_topic_name,
                consumer_group_id=consumer_group_id,
                kafkacrypto=self.__producer_group.kafkacrypto,
                treat_undecryptable_as_plaintext=treat_undecryptable_as_plaintext,
                max_wait_per_decrypt=max_wait_per_decrypt,
                max_initial_wait_per_decrypt=max_initial_wait_per_decrypt,
                logger=self.logger,
            )

    def get_new_subscribed_consumer(self, *, restart_at_beginning=False, **kwargs):
        """
        Return a new Consumer, subscribed to the topic and with the shared group ID.
        Call this function from a child thread to get thread-independent Consumers.

        Note: This function just creates and subscribes the Consumer. Polling it, closing
        it, and everything else must be handled by whatever calls this function.

        :param restart_at_beginning: if True, the new Consumer will start reading partitions
            from the earliest messages available, regardless of Consumer group ID and
            auto.offset.reset values. Useful when re-reading messages.
        :type restart_at_beginning: bool, optional
        :param kwargs: other keyword arguments are passed to the
            :class:`~OpenMSIStreamConsumer` constructor method
        :type kwargs: dict

        :return: a Consumer created using the configs set in the constructor/from `kwargs`,
            subscribed to the topic
        :rtype: :class:`~OpenMSIStreamConsumer`
        """
        if self.__consumer_group is None:
            return None
        return self.__consumer_group.get_new_subscribed_consumer(
            restart_at_beginning=restart_at_beginning, **kwargs
        )

    def get_new_producer(self, **kwargs):
        """
        Return a new :class:`~OpenMSIStreamProducer` object.
        Call this function from a child thread to get thread-independent Producers.
        Note: this function just creates the Producer; closing it etc. must be handled by
        whatever calls this function.

        :return: a Producer created using the config set in the constructor
        :rtype: :class:`~OpenMSIStreamProducer`
        """
        return self.__producer_group.get_new_producer(**kwargs)

    def close(self):
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        """
        if self.__consumer_group is not None:
            self.__consumer_group.close()
        self.__producer_group.close()

    @classmethod
    def get_command_line_arguments(cls):
        """
        Anything extending this class should be able to access the
        "treat_undecryptable_as_plaintext" flag and the
        "max(_initial)_wait_per_decrypt" settings
        """
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "config",
            "topic_name",
            "consumer_topic_name",
            "consumer_group_id",
            "producer_topic_name",
            "treat_undecryptable_as_plaintext",
            "max_wait_per_decrypt",
            "max_initial_wait_per_decrypt",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        args = [
            *superargs,
            parsed_args.config,
        ]
        kwargs = {
            **superkwargs,
            "consumer_topic_name": parsed_args.consumer_topic_name,
            "consumer_group_id": parsed_args.consumer_group_id,
            "producer_topic_name": parsed_args.producer_topic_name,
            "treat_undecryptable_as_plaintext": parsed_args.treat_undecryptable_as_plaintext,
            "max_wait_per_decrypt": parsed_args.max_wait_per_decrypt,
            "max_initial_wait_per_decrypt": parsed_args.max_initial_wait_per_decrypt,
        }
        return args, kwargs



" Various Producibles for main program/base class log messages "

# imports
import json
from time import time
from ..kafka_wrapper import Producible
from .log_handler import LoggingHandler


class LogProducible(Producible):
    """A Producible for log messages. The key is a program ID and the value
    are all of the logs (as provided by handler)
    """

    def __init__(self, program_id):
        self.__program_id = program_id

    @property
    def msg_key(self):
        return f"{self.__program_id}_log"

    @property
    def msg_value(self):
        value_dict = {
            "timestamp": time(),
            "messages": LoggingHandler.get_messages(self.__program_id),
        }
        return json.dumps(value_dict)



" Various message helpers "


def get_message_length(msg):
    """
    returns length of a message
    """
    #
    # This accounting seems imprecise and incomplete.
    #
    keylen = 0
    vallen = 0
    totlen = 0
    if hasattr(msg, "key"):
        try:
            keylen = len(bytes(msg.key))
        except TypeError:
            try:
                keylen = len(msg.key)
            except TypeError:
                pass
    if hasattr(msg, "value"):
        try:
            vallen = len(bytes(msg.value))
        except TypeError:
            try:
                vallen = len(msg.value)
            except TypeError:
                pass
    try:
        totlen = len(bytes(msg))
    except TypeError:
        try:
            totlen = len(msg)
        except TypeError:
            pass
    return max(keylen + vallen, totlen)



"""Small utility class to simplify producing messages with standardized callbacks"""

# imports
from abc import ABC, abstractmethod


class Producible(ABC):
    """
    Small utility class for anything that can be Produced as a message to a topic
    """

    @property
    @abstractmethod
    def msg_key(self):
        """
        The key of the object when it's represented as a message.
        This can be something that needs to be serialized still

        Not implemented in base class
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def msg_value(self):
        """
        The value of the object when it's represented as a message.
        This can be something that needs to be serialized still

        Not implemented in base class
        """
        raise NotImplementedError

    @property
    def callback_kwargs(self):
        """
        A dictionary of keyword arguments that should be sent to the callback function
        for Producers producing messages of this Producible

        Empty for the base class
        """
        return {}

    def get_log_msg(self, print_every=None):  # pylint: disable=unused-argument
        """
        Given some (optional) "print_every" variable, return the string that should be
        logged (at debug level) for this message.
        If "None" is returned (the default) nothing will be logged.
        """
        return None



" Global handler of log messages, to allow self-production of own (complete) logs "

# imports
from logging import Handler
from warnings import formatwarning
from threading import Lock

# from sys import stderr


class LoggingHandlerClass(Handler):
    """A thread-safe, global, handler for log messages.
    Implemented as a ring buffer to avoid unbounded resource usage
    in case of non-production or rapid log message generation rate.
    """

    def __init__(self, max_messages=128):
        # we set max_messages small here, and increase when a log user appears
        # this should be at least as large as the number of log messages
        # expected from program start until the first log producer is created
        super().__init__()
        self.__buffer = [None] * max_messages
        self.__buffer_size = max_messages
        self.__write_pointer = [0, 0]  # index, generation (# buffer wraps)
        self.__read_pointers = {}
        self.__write_lock = Lock()  # protects write_pointer and buffer
        self.__read_lock = Lock()  # protects read_pointers

    def set_max_messages(self, max_messages):
        "Change log ring buffer size"
        with self.__read_lock:
            with self.__write_lock:
                # Try to preserve logs even when shrinking
                lag_read_pointer = self.__write_pointer[0]
                for rpn in self.__read_pointers:  # pylint: disable=C0206
                    if self.__read_pointers[rpn][0] <= self.__write_pointer[0]:
                        if self.__read_pointers[rpn][1] == self.__write_pointer[1]:
                            nlrp = self.__read_pointers[rpn][0]
                        else:
                            # This reader is really far behind and already wrapped
                            nlrp = (self.__write_pointer[0] + 1) - max_messages
                    else:
                        if self.__read_pointers[rpn][1] == (self.__write_pointer[1] - 1):
                            nlrp = self.__read_pointers[rpn][0] - max_messages
                        else:
                            # This reader is really far behind and already wrapped
                            nlrp = (self.__write_pointer[0] + 1) - max_messages
                    # store offset for later computation of new start point
                    self.__read_pointers[rpn] = [nlrp, 0]
                    if nlrp < lag_read_pointer:
                        lag_read_pointer = nlrp
                # New buffer covers past X (partly) unread events,
                # with X <= max_messsages, and max_messages-X new
                # slots.
                lag_read_pointer = max(
                    self.__write_pointer[0] - max_messages, lag_read_pointer
                )
                old_buffer = self.__buffer
                self.__buffer = []
                if lag_read_pointer < 0:
                    self.__buffer += old_buffer[lag_read_pointer:]
                if lag_read_pointer < self.__write_pointer[0]:
                    self.__buffer += old_buffer[
                        max(lag_read_pointer, 0) : self.__write_pointer[0]
                    ]
                self.__write_pointer = [len(self.__buffer), 0]
                if self.__write_pointer[0] < max_messages:
                    self.__buffer += [None] * (max_messages - self.__write_pointer[0])
                else:
                    # map index max_messages back to first slot
                    self.__write_pointer = [0, 0]
                self.__buffer_size = max_messages
                # update read pointers
                for rpn in self.__read_pointers:  # pylint: disable=C0206
                    self.__read_pointers[rpn][0] -= lag_read_pointer
                    self.__read_pointers[rpn][1] = 0
                    if self.__read_pointers[rpn][0] < 0:
                        # This reader will have missed messages
                        self.__read_pointers[rpn][0] = 0

    def emit(self, record):
        with self.__write_lock:
            self.__buffer[self.__write_pointer[0]] = self.format(record)
            self.__write_pointer[0] += 1
            if self.__write_pointer[0] == self.__buffer_size:
                self.__write_pointer[1] += 1
                self.__write_pointer[0] = 0

    # We disable pylint here due as we must match specified showwarning call signature
    def showwarning(
        self, message, category, filename, lineno, file=None, line=None
    ):  # pylint: disable=W0613
        "Python warnings showwarning implementation"
        msg = formatwarning(message, category, filename, lineno, line)
        with self.__write_lock:
            self.__buffer[self.__write_pointer[0]] = msg
            self.__write_pointer[0] += 1
            if self.__write_pointer[0] == self.__buffer_size:
                self.__write_pointer[1] += 1
                self.__write_pointer[0] = 0
        # should we still do output as default showwarning would?
        # if file is None:
        #     file = stderr
        # print(msg, file=file, flush=True)

    def get_messages(self, reader_id="default"):
        "Get unread messages for given reader"
        start_pos = None
        end_pos = None
        rv = []
        with self.__read_lock:
            if reader_id not in self.__read_pointers:
                self.__read_pointers[reader_id] = [
                    0,
                    -1,
                ]  # index, generation=-1 (= current writer)
            start_pos = self.__read_pointers[reader_id]
        with self.__write_lock:
            end_pos = self.__write_pointer
            if start_pos[1] == -1:
                start_pos[1] = end_pos[1]
            if start_pos[0] < end_pos[0] and start_pos[1] == end_pos[1]:
                rv = self.__buffer[start_pos[0] : end_pos[0]].copy()
            elif start_pos[0] > end_pos[0] and start_pos[1] == (end_pos[1] - 1):
                rv = (
                    self.__buffer[start_pos[0] :].copy()
                    + self.__buffer[0 : end_pos[0]].copy()
                )
            elif start_pos[1] < (end_pos[1] - 1):
                rv = (
                    self.__buffer[end_pos[0] :].copy()
                    + self.__buffer[0 : end_pos[0]].copy()
                )
            # start_pos = end_pos is empty list
        with self.__read_lock:
            self.__read_pointers[reader_id] = end_pos.copy()
        return rv

    def close(self):
        super().close()


# The actual, single, global instance of LoggingHandlerClass
LoggingHandler = LoggingHandlerClass()
